#ifndef __PTRACE_MODULE_H
#define __PTRACE_MODULE_H

#include <sys/types.h>

#include <complex.h>
#include <stdbool.h>
#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif

/* Forward declaration. */
struct bt_variable;

typedef struct module pmodule_t;
typedef bool pmodule_load_fn_t(pmodule_t *);
typedef void pmodule_reset_fn_t(pmodule_t *);
typedef void pmodule_unload_fn_t(pmodule_t *);

struct pmodule_internal {
	const char *id;
	unsigned int version;
	pmodule_load_fn_t *load;
	pmodule_reset_fn_t *reset;
	pmodule_unload_fn_t *unload;
	int priority;
};

/* ABI Counter */
#define PMODULE_VERSION 16

/**
 * Define your module.
 * @name id string for your module.
 * @load_func registration function for your module
 *    -invoked before initial snapshot
 * @reset function to reset any state
 *    -invoked before each subsequent snapshot
 * @unload unregistration function for your module
 *    -invoked after last snapshot
 *
 * NOTE:
 *	A call to reset always precedes a call to unload.
 *	.unload is invoked during ptrace breakdown after the last snapshot is completed.
 *	.reset is invoked at the end of each iteration.
 *	A load function is required.
 *
 */
#define PMODULE_DEFINE(n, loadf, resetf, unloadf)	\
	static pmodule_load_fn_t loadf;			\
	const struct pmodule_internal __ptmod_obj = {	\
		.id = n,				\
		.version = PMODULE_VERSION,		\
		.load = loadf,				\
		.reset = resetf,			\
		.unload = unloadf			\
	};

/*
 * This may be used to allow for static linkage and bundling of pmodule
 * objects.
 */
#define PMODULE_STATIC_DEFINE(N, loadf, resetf, unloadf, P)		\
	static pmodule_load_fn_t loadf;					\
	static const struct pmodule_internal _ptmod_obj_##N		\
	= {								\
		.id = #N,						\
		.load = (loadf),					\
		.reset = (resetf),					\
		.unload = (unloadf),					\
		.priority = (P)						\
	};								\
	static void							\
	__attribute__((constructor))					\
	__attribute__((unused))						\
	pmodule_register_##N(void)					\
	{								\
		ptrace_module_register(&_ptmod_obj_##N);		\
		return;							\
	}

/*
 * Config API
 */

/**
 * Retrieve option passed to module via command line.
 * @param1 module context
 * @param2 key for option
 */
const char *pmodule_config_cli_option(pmodule_t *, const char *);

/**
 * Retrieve asset path for module. ($PREFIX_PATH/$MODULE_NAME/)
 * @param1 module context
 * @param2 pointer to storage for path
 * @param3 size of storage of param2
 */
int pmodule_config_asset_path(pmodule_t *, char *, size_t);
/**
 * Get the path of the module's config file (specified via cmd line or default)
 * @param1 module context
 * @param2 buffer to store config file path
 * @param3 size of param2
 *
 * @return true on success, false otherwise
 */
bool pmodule_config_module_config_file(pmodule_t *, char *, size_t);
/**
 * Utility function to detect comment lines in module configuration files
 * @param1 ptr to line
 */
bool pmodule_config_is_comment(const char *);
/**
 * Utility function to eat up whitespace
 * @param1 ptr to string
 * @param2 maximum scan length
 */
bool pmodule_config_chomp(const char **, size_t);

/**
 * Retrieve json object from main config
 * @param1 module context
 */
struct json_object;
struct json_object *pmodule_config_json_config(pmodule_t *);

/**
 * Match API
 * Note: memory for strings stored in the match object is maintained
 * by the owner of the match object.
 * See pmodule_match_create(), pmodule_match_init(), pmodule_match_destroy(),
 * and pmodule_match_free().
 */

typedef struct pmodule_match pmodule_match_t;

/**
 * Determines whether a string match should be exact or a substring match.
 */
enum pmodule_match_sel {
	PMODULE_MATCH_EXACT = 0,
	PMODULE_MATCH_SUBSTR,
	PMODULE_MATCH_MAX
};

#define PTRACE_VARIABLE_S_RESOLVED	(1U << 0)
#define PTRACE_VARIABLE_S_FAULT		(1U << 1)
#define PTRACE_VARIABLE_S_ANNOTATED	(1U << 2)
#define PTRACE_VARIABLE_S_SKIP		(1U << 3)

enum ptrace_variable_type {
	PTRACE_VARIABLE_TYPE_LL = 0,	/* long long */
	PTRACE_VARIABLE_TYPE_ULL,	/* unsigned long long */
	PTRACE_VARIABLE_TYPE_FLT,	/* float */
	PTRACE_VARIABLE_TYPE_DBL,	/* double */
	PTRACE_VARIABLE_TYPE_LDBL,	/* long double */
	PTRACE_VARIABLE_TYPE_CFLT,	/* complex float */
	PTRACE_VARIABLE_TYPE_CDBL,	/* complex double */
	PTRACE_VARIABLE_TYPE_CLDBL,	/* complex long double */
	PTRACE_VARIABLE_TYPE_STRING,	/* string */
	PTRACE_VARIABLE_TYPE_ENUM,	/* enumerator */
	PTRACE_VARIABLE_TYPE_TUPLE,	/* tuple array */
	PTRACE_VARIABLE_TYPE_ARRAY,	/* single-type array */
	PTRACE_VARIABLE_TYPE_BOTTOM,	/* no type */
	PTRACE_VARIABLE_TYPE_FUNCTION,	/* function */
	PTRACE_VARIABLE_TYPE_REFERENCE,	/* reference */
	PTRACE_VARIABLE_TYPE_MAX
};

/**
 * Create a new match object.
 */
pmodule_match_t *pmodule_match_create(void);
/**
 * Initialize a match object to its base (empty) state.
 */
void pmodule_match_init(pmodule_match_t *m);
/**
 * Match objects are copied during pmodule_*_event_add()
 * Destroy the match object once you are done with it
 * to cleanup memory allocated internally.
 * Note: This does not free the owning match object.
 */
void pmodule_match_destroy(pmodule_match_t *m);
/**
 * Free a match object created with pmodule_match_create().
 */
void pmodule_match_free(pmodule_match_t *m);
/**
 * Add an object file criterion.
 * @param 1 the match object to modify
 * @param 2 the object file name
 * @param 3 exact or substring match
 * @return true if match criterion was added successfully, false otherwise
 */
bool pmodule_match_add_object(pmodule_match_t *m, const char *obj_name, enum pmodule_match_sel s);
/**
 * Add a source file criterion.
 * @param 1 the match object to modify
 * @param 2 the source file name
 * @param 3 exact or substring match
 * @return true if match criterion was added successfully, false otherwise
 */
bool pmodule_match_add_file(pmodule_match_t *m, const char *file_name, enum pmodule_match_sel s);
/**
 * Add a frame symbol criterion.
 * @param 1 the match object to modify
 * @param 2 the frame symbol name
 * @param 3 exact or substring match
 * @return true if match criterion was added successfully, false otherwise
 */
bool pmodule_match_add_frame_sym(pmodule_match_t *m, const char *sym, enum pmodule_match_sel s);
/**
 * Add a variable type name criterion.
 * @param 1 the match object to modify
 * @param 2 the variable type name
 * @param 3 exact or substring match
 * @return true if match criterion was added successfully, false otherwise
 */
bool pmodule_match_add_variable_base_type(pmodule_match_t *m, const char *type, enum pmodule_match_sel s);
/**
 * Add a variable name criterion.
 * @param 1 the match object to modify
 * @param 2 the variable name
 * @param 3 exact or substring match
 * @return true if match criterion was added successfully, false otherwise
 */
bool pmodule_match_add_variable_name(pmodule_match_t *m, const char *var_name, enum pmodule_match_sel s);
/**
 * Add a variable ptrace type criterion.
 * @param 1 the match object to modify
 * @param 2 the ptrace type name
 * @param 3 exact or substring match
 * @return true if match criterion was added successfully, false otherwise
 */
bool pmodule_match_add_variable_ptrace_type(pmodule_match_t *m, enum ptrace_variable_type);

/**
 * Fault is true iff:
 *	variables: the containing frame has fault set.
 *	frames: the faulting event happens in this frame.
 *	thread: the thread contains a frame with fault set.
 *
 * If fault isn't set within a pmodule_match object, then
 * the pmodule_match object will match on all frames
 * (ignoring fault.)
 *
 * If you want to match on all [variables, frames] from the
 * faulting thread, use pmodule_match_thread_fault_set
 */
void pmodule_match_fault_set(pmodule_match_t *m);

/**
 * Match if the containing thread is faulting
 */
void pmodule_match_thread_fault_set(pmodule_match_t *m);

/**
 * Event listen API
 * Execute the module desc by 'ctx' when an 'event' occurs.
 * Events/Phases:
 *	variable:   When a new variable has been extracted.
 *	frame:      When a new frame has been extracted.
 *	thread:     When a new thread has been extracted
 *	preattach:  Before ptrace attaches to the process.
 *	postattach: After ptrace detaches from the process.
 *	attach:     Right after attachment, before extraction.
 */

typedef struct variable ptrace_variable_t;
typedef struct frame ptrace_frame_t;
typedef struct ptrace_thread ptrace_thread_t;
typedef struct backtrace ptrace_backtrace_t;

struct bt_variable;
typedef void pmodule_variable_event_cb_t(ptrace_variable_t *,
    struct bt_variable *);
typedef void pmodule_frame_event_cb_t(ptrace_frame_t *);
typedef void pmodule_thread_event_cb_t(ptrace_thread_t *);
typedef void pmodule_preattach_event_cb_t(ptrace_backtrace_t *);
typedef void pmodule_postattach_event_cb_t(ptrace_backtrace_t *);
typedef void pmodule_attach_event_cb_t(ptrace_backtrace_t *);

bool pmodule_variable_event_add(pmodule_t *ctx,
    pmodule_match_t *m,
    pmodule_variable_event_cb_t *);

bool pmodule_frame_event_add(pmodule_t *ctx,
    pmodule_match_t *m,
    pmodule_frame_event_cb_t *);

bool pmodule_thread_event_add(pmodule_t *ctx,
    pmodule_match_t *m,
    pmodule_thread_event_cb_t *);

bool pmodule_preattach_event_add(pmodule_t *ctx,
    pmodule_match_t *m,
    pmodule_preattach_event_cb_t *);

bool pmodule_postattach_event_add(pmodule_t *ctx,
    pmodule_match_t *m,
    pmodule_postattach_event_cb_t *);

bool pmodule_attach_event_add(pmodule_t *ctx,
    pmodule_match_t *m,
    pmodule_attach_event_cb_t *);

/**
 * Annotation API
 * Variables, frames, threads, and the overarching backtrace object
 * may be annotated with additional data. All annotation types but
 * PAT_JSON(_UNSAFE) will appear as text attached to the specified object;
 * PAT_JSON(_UNSAFE) will result in the specified text being parsed as a
 * json object to be attached to the specified object.
 * PAT_JSON_PP will be parsed as a json object but will also be displayed
 * with higher priority as a variable's "true" output (e.g. this annotation
 * will be displayed as a child of the variable before any of the variable's
 * actual child members/variables).
 *
 * JSON annotation formatting:
 * Hydra will parse structured JSON annotations as follows:
 *
 * "json": {
 * 	"context" : "Title",
 * 	"key1": "value",
 * 	"key2": "value",
 * 	"key3": {
 * 		"key1": "value",
 * 		"key2": "value"
 * 	}
 * }
 *
 * Rendered output (in Hydra):
 *
 * - Title
 *	key1	value
 *	key2	value
 *    - key3
 *		key1	value
 *		key2	value
 *
 * "json" must be the overall key to the annotation object, as the annotation
 * will be embedded into another json object.
 *
 * "context" is the title of this particular section; the rest of the output
 * will appear in a hierarchy under the "context" title.
 *
 * All other fields are key value pairs. Any values which are objects themselves
 * will increase the hierarchy level by one until that object ends. Thus, output
 * may be structured into a tree view.
 */

enum pmodule_annotation_type {
	PAT_COMMENT = 0,
	PAT_WARNING,
	PAT_CRITICAL,
	PAT_JSON,
	PAT_JSON_UNSAFE,
	PAT_JSON_PP,
	PAT_MAX
};

/*
 * This function directly takes a JSON key-value pair where key is "json" for
 * embedding. No additional validation is provided and it is possible to write
 * invalid JSON to a snapshot. Use with caution and only if JSON validation
 * is a significant bottleneck.
 */
#define pmodule_annotate_backtrace_unsafe(A, B, ...)			\
	_pmodule_annotate_backtrace_unsafe((A),				\
	    "{\"object_id\":0, \"type\":\"JSON\"," B "}", ##__VA_ARGS__)

bool _pmodule_annotate_backtrace_unsafe(ptrace_backtrace_t *,
    const char *, ...) __attribute__((format(printf, 2, 3)));

bool pmodule_annotate_backtrace(ptrace_backtrace_t *,
    enum pmodule_annotation_type,
    const char *,
    ...)
    __attribute__((format(printf, 3, 4)));

bool pmodule_annotate_thread(ptrace_thread_t *,
    enum pmodule_annotation_type,
    const char *,
    ...)
    __attribute__((format(printf, 3, 4)));

bool pmodule_annotate_frame(ptrace_frame_t *,
    enum pmodule_annotation_type,
    const char *,
    ...)
    __attribute__((format(printf, 3, 4)));

bool pmodule_annotate_variable(ptrace_variable_t *,
    enum pmodule_annotation_type,
    const char *,
    ...)
    __attribute__((format(printf, 3, 4)));

/*
 * This function directly takes a JSON key-value pair where key is "json" for
 * embedding. No additional validation is provided and it is possible to write
 * invalid JSON to a snapshot. Use with caution and only if JSON validation
 * is a significant bottleneck.
 */
#define pmodule_annotate_variable_unsafe(A, B, ...) ({			\
	ptrace_variable_t *_pmodule_v = (A);				\
	_pmodule_annotate_variable_unsafe(_pmodule_v,			\
	    "{\"object_id\":\"%llu\", \"type\":\"JSON\"," B "}",	\
	    pmodule_variable_object_id(_pmodule_v), ##__VA_ARGS__);	\
})

bool _pmodule_annotate_variable_unsafe(ptrace_variable_t *,
    const char *,
    ...)
    __attribute__((format(printf, 2, 3)));

/* Global API */

/**
 * Returns the address of the specified symbol.
 * @param 1 the symbol to lookup
 * @param 2 the object file for the lookup; may be NULL
 * @param 3 storage for the symbol's address
 * @return true if found, false otherwise
 */
bool pmodule_symbol_address(const char *, const char *,
    unsigned long long *);

/**
 * Returns the symbol associated with the specified address.
 * @param 1 the address to lookup
 * @return the symbol of the address if found, NULL otherwise
 */
const char *pmodule_address_symbol(unsigned long long);

/**
 * Reads the specified length number of bytes from the location of
 * the supplied symbol into the supplied buffer of the same specified
 * length.
 * @param 1 the symbol to lookup
 * @param 2 the object file for the lookup; may be NULL
 * @param 3 the buffer in which to store the data
 * @param 4 the maximum number of bytes to read
 * @return the number of bytes read, or -1 on error
 */
ssize_t pmodule_symbol_memory_read(const char *, const char *, void *, size_t);

/**
 * Reads the specified length number of bytes from the location of
 * the supplied address into the supplied buffer of the same specified
 * length.
 * @param 1 the address to lookup
 * @param 2 the buffer in which to store the data
 * @param 3 the maximum number of bytes to read
 * @return the number of bytes read, or -1 on error
 */
ssize_t pmodule_address_memory_read(unsigned long long, void *, size_t);

/**
 * Extracts a string value from a symbol of up to limit bytes or first
 * null character.
 * @param 1 the symbol to lookup
 * @param 2 the object file for the lookup; may be NULL
 * @param 3 the buffer in which to store the string
 * @param 4 the maximum number of bytes to read
 * @return the number of bytes read, or -1 on error
 */
ssize_t pmodule_symbol_string_read(const char *, const char *, char *, size_t);

/**
 * Extracts a string value from an address of up to limit bytes or first
 * null character.
 * @param 1 the address to lookup
 * @param 2 the buffer in which to store the string
 * @param 3 the maximum number of bytes to read
 * @return the number of bytes read, or -1 on error
 */
ssize_t pmodule_address_string_read(unsigned long long, char *, size_t);

/**
 * Returns the parameter from the specified frame with the
 * specified name.
 * @param 1 the frame to search for the parameter
 * @param 2 the name of the parameter
 * @return a reference to a variable from the specified frame if found, NULL otherwise
 */
ptrace_variable_t *pmodule_frame_param(ptrace_frame_t *, const char *);

/**
 * Returns the ptrace_variable matching the specified global symbol.
 * @param 1 the name of the global variable
 * @return a reference to a variable if found, NULL otherwise
 */
ptrace_variable_t *pmodule_symbol_variable(const char *);

/**
 * Iteration API
 *
 * Iterate over threads, frames, and variables satisfying the specified
 * match criteria (NULL is a valid match object).
 *
 * Callback return params:
 *	true == stop iteration.
 *	false == contnue;
 */

typedef bool pmodule_scan_threads_cb_t(ptrace_thread_t *, void *);
typedef bool pmodule_scan_frames_cb_t(ptrace_frame_t *, void *);
typedef bool pmodule_scan_variables_cb_t(ptrace_variable_t *, void *);

bool pmodule_backtrace_scan_threads(ptrace_backtrace_t *, pmodule_match_t *,
    pmodule_scan_threads_cb_t *, void *);
bool pmodule_thread_scan_frames(ptrace_thread_t *, pmodule_match_t *,
    pmodule_scan_frames_cb_t, void *);
bool pmodule_frame_scan_variables(ptrace_frame_t *, pmodule_match_t *,
    pmodule_scan_variables_cb_t *, void *);

/*
 * Pmodule_mm API (memmap)
 */

typedef struct mm_address ptrace_mm_address_t;

/**
 * The return paramter is a ptrace_mm_address object,
 * which contains a list of variable aliases for a
 * specific memory address.
 */
ptrace_mm_address_t *pmodule_mm_address_get(unsigned long long);

typedef bool pmodule_address_iterate_cb_t(ptrace_variable_t *, void *);
/**
 * Iterate over the list of variable aliases for
 * a specific memory address. A loop version of this
 * exists as PTRACE_MM_ADDRESS_FOREACH
 *
 * @param 3 stop iteration after first match. A
 * memory map node stores back refs to all variables that
 * reference it. In some cases (to find ancestors) you only
 * want a single variable that references the mm_node.
 */
bool pmodule_mm_address_iterate(pmodule_match_t *,
    ptrace_mm_address_t *,
    bool,
    pmodule_address_iterate_cb_t *,
    void *);

typedef bool pmodule_ancestors_iterate_cb_t(ptrace_variable_t *, void *);
/**
 * Ancestors is a list of variables that reference
 * the ptif_mm_variable contained within argument 2.
 * There is only a single instance of a ptif_mm_variable
 * for each (address, ptrace_type) tuple.
 *
 * @param 3 preempt iteration after first match. A memory
 * map node stores back refs to all variables that reference it.
 * In some cases, you only want a single instance of a variable
 * that is an ancestor.
 */
bool pmodule_mm_ancestors_iterate(pmodule_match_t *,
    ptrace_variable_t *,
    bool,
    pmodule_ancestors_iterate_cb_t *,
    void *);

/* Logging API */
enum pmodule_log_level {
	PMODULE_DEBUG = 0,
	PMODULE_INFO,
	PMODULE_WARNING,
	PMODULE_ERR,
	PMODULE_MAX
};
#define PMODULE_LOG(l, m, ...)					\
	pmodule_log(l, "[%s:%d:%s] "m,				\
	    __FILE__, __LINE__, __FUNCTION__, ##__VA_ARGS__)
/*
 * PMODULE_LOG_S is the same as PMODULE_LOG except it requests to serialize the
 * message to snapshot regardless of the log level
 */
#define PMODULE_LOG_SERIALIZE (1u << 31)
#define PMODULE_LOG_S(l, m, ...)		\
	PMODULE_LOG((l) | PMODULE_LOG_SERIALIZE, m, ##__VA_ARGS__)

void pmodule_log(enum pmodule_log_level, const char *, ...)
    __attribute__((format(printf, 2, 3)));
bool pmodule_log_level_set(enum pmodule_log_level);

/* String Functions */
char *pmodule_strerror(int errnum);
/* Ensures null terminated string */
char *pmodule_strncpy(char *, const char *, size_t);

/* Signal */
typedef struct ptif_siginfo ptif_siginfo_t;

bool pmodule_signal_valid(ptif_siginfo_t *);
const char *pmodule_signal_string(ptif_siginfo_t *);
const char *pmodule_signal_reason(ptif_siginfo_t *, bool *k);
bool pmodule_signal_address_populated(ptif_siginfo_t *);
int pmodule_signal_num(ptif_siginfo_t *);
int pmodule_signal_code(ptif_siginfo_t *);
unsigned long long pmodule_signal_address(ptif_siginfo_t *);
ptif_siginfo_t *pmodule_frame_siginfo(ptrace_frame_t *);

/* Object API */

/* Variables */

/**
 * @brief Process a new variable through the module hooks.
 *
 * This is to be used when adding new variables in modules.
 */
void pmodule_variable_dispatch(ptrace_variable_t *, struct bt_variable *);

ptrace_variable_t *pmodule_ptrace_variable_create(ptrace_variable_t *pv,
    unsigned int depth, size_t offset);
bool pmodule_variable_value(ptrace_variable_t *pv,
    struct bt_variable *bv, bool);

unsigned int pmodule_variable_depth(ptrace_variable_t *pv);
unsigned int pmodule_variable_state(ptrace_variable_t *pv);
void pmodule_variable_state_set(ptrace_variable_t *pv, unsigned int);

unsigned long long pmodule_variable_address(ptrace_variable_t *v);
/**
 * Auxiliary symbol data associated with the variable, depending on
 * its type and availability.
 * @return Returns the following according to the variable type:
 * string - the C-string value pointed to by the variable
 * enum - The label of the enum
 * function - The name of the function
 * reference - The name of the pointed-to variable
 * NULL if there is no symbol data available
 */
const char *pmodule_variable_symbol(ptrace_variable_t *v);
enum ptrace_variable_type pmodule_variable_type(ptrace_variable_t *v);
/**
 * The parent of the specified variable (relevant if this variable
 * is a member of a struct or array).
 */
ptrace_variable_t *pmodule_variable_owner(ptrace_variable_t *v);
unsigned long long pmodule_variable_object_id(ptrace_variable_t *v);
bool pmodule_variable_has_annotations(ptrace_variable_t *v);
/**
 * The frame which contains this variable.
 */
ptrace_frame_t *pmodule_variable_frame(ptrace_variable_t *v);
/**
 * The overarching backtrace object which contains this variable.
 */
ptrace_backtrace_t *pmodule_variable_backtrace(ptrace_variable_t *v);
/**
 * The frame parameter index of the specified variable.
 * @return the offset of the variable (offsets start from 0) if variable
 * is a frame parameter; -1 if not
 */
long long int pmodule_variable_fprm_offset(ptrace_variable_t *v);
bool pmodule_variable_resolved(ptrace_variable_t *v);
/**
 * The name of the variable.
 */
const char *pmodule_variable_name(ptrace_variable_t *v);
/**
 * The demangled, C-style variable type name.
 */
const char *pmodule_variable_type_name(ptrace_variable_t *v);

/*
 * Value accessors - the following will extract the appropriate values
 * from the specified variable.
 */
unsigned long long pmodule_variable_ref(ptrace_variable_t *v);
double complex pmodule_variable_cdbl(ptrace_variable_t *v);
long double complex pmodule_variable_cldbl(ptrace_variable_t *v);
float complex pmodule_variable_cflt(ptrace_variable_t *v);
double pmodule_variable_dbl(ptrace_variable_t *v);
long double pmodule_variable_ldbl(ptrace_variable_t *v);
float pmodule_variable_flt(ptrace_variable_t *v);
unsigned long long pmodule_variable_ull(ptrace_variable_t *v);
long long pmodule_variable_ll(ptrace_variable_t *v);

/**
 * Copies the string contained in the specified string variable to the
 * specified buffer, up to a maximum of limit bytes.
 * @param 1 output buffer
 * @param 2 variable from which to copy
 * @param 3 maximum number of bytes to copy
 * @return number of bytes copied; -1 on failure
 */
ssize_t pmodule_variable_strncpy(char *buffer, ptrace_variable_t *v, size_t limit);
bool pmodule_variable_is_string(ptrace_variable_t *v);
const char *pmodule_variable_base_type(ptrace_variable_t *v);
/**
 * Checks a variable's type.
 * @return true if the specified variable is of the specified type
 */
bool pmodule_variable_check_type(ptrace_variable_t *, enum ptrace_variable_type);

/* Frames */

/**
 * The instruction of the specified frame.
 */
const char *pmodule_frame_instruction(ptrace_frame_t *fr);
/**
 * The source file of the specified frame.
 */
const char *pmodule_frame_path(ptrace_frame_t *);
/**
 * The demangled symbol of the specified frame.
 */
const char *pmodule_frame_symbol_demangle(ptrace_frame_t *);
/**
 * The status of the frame's line information.
 */
bool pmodule_frame_line_resolved(ptrace_frame_t *);
/**
 * The object file of the specified frame.
 */
const char *pmodule_frame_object_file(ptrace_frame_t *);
/**
 * The backtrace object containing the specified frame.
 */
ptrace_backtrace_t *pmodule_frame_backtrace(ptrace_frame_t *);
/**
 * The thread object containing the specified frame.
 */
ptrace_thread_t *pmodule_frame_thread(ptrace_frame_t *);
unsigned long long pmodule_frame_object_id(ptrace_frame_t *);
/**
 * The current source line of the specified frame.
 */
unsigned long long pmodule_frame_line(ptrace_frame_t *);
/**
 * The declaration line of the specified frame.
 */
unsigned long pmodule_frame_decl_line(ptrace_frame_t *);
/**
 * The variable with the indicated frame parameter offset.
 * @param 1 the frame to search
 * @param 2 the parameter offset (begins at 0)
 */
ptrace_variable_t *pmodule_frame_fprm(ptrace_frame_t *, unsigned long);

/* Thread API */
bool pmodule_thread_is_fault(ptrace_thread_t *);
/**
 * @return true if the thread is a userspace thread; false if the thread is a
 * kernel thread
 */
bool pmodule_thread_is_userspace(ptrace_thread_t *);
/**
 * The backtrace object containing the specified thread.
 */
ptrace_backtrace_t *pmodule_thread_backtrace(ptrace_thread_t *);
unsigned long long pmodule_thread_object_id(ptrace_thread_t *);
/**
 * @return number of frames in the specified thread
 */
unsigned long pmodule_thread_n_frames(ptrace_thread_t *);
/**
 * @return number of kernel frames in the specified thread
 */
unsigned long pmodule_thread_n_k_frames(ptrace_thread_t *);
/**
 * Sets a thread's name. This will overwrite any previously detected
 * threadname.
 * @param 1 the thread to modify
 * @param 2 the new name to set
 * @return true if name was successfully changed, false otherwise
 */
bool pmodule_thread_name_set(ptrace_thread_t *, const char *);

/* Backtrace API */
/**
 * Note: for tags, kv_*, the key/tag is required to be unique.  If duplicate
 * keys are added, the last one shall be stored.
 */

/**
 * @return the global backtrace object for the current ptrace run
 */
ptrace_backtrace_t *pmodule_backtrace_get(void);
/**
 * Adds a tag to the backtrace.
 * @param 2 the tagname to add
 * @return true if the tag was successfully added, false otherwise
 */
bool pmodule_backtrace_add_tag(ptrace_backtrace_t *, const char *);
/**
 * Adds a key-value int to the backtrace.
 * @param 2 the key string
 * @param 3 the value int
 * @return true if the kv-int was successfully added, false otherwise
 */
bool pmodule_backtrace_add_kv_int(ptrace_backtrace_t *,
    const char *,
    uint64_t);
/**
 * Adds a key-value string to the backtrace.
 * @param 2 the key string
 * @param 3 the value string
 * @return true if the kv-string was successfully added, false otherwise
 */
bool pmodule_backtrace_add_kv_string(ptrace_backtrace_t *,
    const char *,
    const char *);
pid_t pmodule_backtrace_pid(ptrace_backtrace_t *);
const char *pmodule_backtrace_application_name(ptrace_backtrace_t *);

struct bt_process;
/**
 * The process structure associated with the backtrace for use with
 * the libbt API.
 */
struct bt_process *pmodule_backtrace_process(ptrace_backtrace_t *);

enum pmodule_process_state {
	PMODULE_STATUS_RUNNING = 0,
	PMODULE_STATUS_SLEEPING,
	PMODULE_STATUS_DISK,
	PMODULE_STATUS_STOPPED,
	PMODULE_STATUS_TRACED,
	PMODULE_STATUS_ZOMBIE,
	PMODULE_STATUS_DEAD,
	PMODULE_STATUS_IDLE,
	PMODULE_STATUS_BLOCKED,
	PMODULE_STATUS_QUEUED,
	PMODULE_STATUS_UNKNOWN,
	PMODULE_STATUS_RUNNABLE,
	PMODULE_STATUS_SYSCALL,
	PMODULE_STATUS_WAITING,
	PMODULE_STATUS_COPYSTACK,
	PMODULE_STATUS_MAX
};

enum pmodule_process_state pmodule_backtrace_process_state(ptrace_backtrace_t *);

/* Classifier API */
typedef struct classifier ptrace_classifier_t;

/*
 * @param1 the backtrace
 * @param2 classifier name
 * @return classifier object on success, NULL otherwise.
 */
ptrace_classifier_t *pmodule_classifier_new(ptrace_backtrace_t *, const char *);

/*
 * Associates object with id specified by parameter 3 to
 * classifier specified by parameter 2
 */
bool pmodule_classifier_ref_object(ptrace_backtrace_t *,
    ptrace_classifier_t *,
    unsigned long long);

/* Label API */
typedef struct label ptrace_label_t;

/*
 * returns NULL on error, label object otherwise.
 * If label does not already exist, adds it to the backtrace.
 */
ptrace_label_t *pmodule_label_new(ptrace_backtrace_t *,
    const char *,
    size_t,
    unsigned char *);
/*
 * Associates object with id specified by parameter 3 to
 * label specified by parameter 2
 */
bool pmodule_label_ref_object(ptrace_backtrace_t *,
    ptrace_label_t *,
    unsigned long long);

/* P(rocess)map API */
typedef struct pmap_entry ptrace_pmap_entry_t;

bool pmodule_pmap_enabled(void);
bool pmodule_pmap_cached(void);
ptrace_pmap_entry_t *pmodule_pmap_find_address(ptrace_backtrace_t *,
    unsigned long long);

enum pmodule_pmap_flag {
	PMODULE_PMAP_F_EX = 0,
	PMODULE_PMAP_F_RD,
	PMODULE_PMAP_F_WR,
	PMODULE_PMAP_F_PR,
	PMODULE_PMAP_F_SH,
	PMODULE_PMAP_F_ALL,
	PMODULE_PMAP_F_MAX
};

#define PMODULE_PMAP_MODE_NONE	0
#define PMODULE_PMAP_MODE_EX	(1U << PMODULE_PMAP_F_EX)
#define PMODULE_PMAP_MODE_RD	(1U << PMODULE_PMAP_F_RD)
#define PMODULE_PMAP_MODE_WR	(1U << PMODULE_PMAP_F_WR)
#define PMODULE_PMAP_MODE_PR	(1U << PMODULE_PMAP_F_PR)
#define PMODULE_PMAP_MODE_SH	(1U << PMODULE_PMAP_F_SH)
#define PMODULE_PMAP_MODE_ALL	(UINT_MAX)
#define PMODULE_PMAP_MODE(flag)	(1U << (flag))

bool pmodule_pmap_permission_flags(ptrace_pmap_entry_t *, unsigned int);
bool pmodule_pmap_is_guard(ptrace_pmap_entry_t *);

/* Synthesis API */

/* Variables */

/**
 * Synthesizes a variable from user-supplied data. In order to appear
 * in a backtrace, synthesized variables must be added to a frame
 * (and if that frame is also synthesized, it in turn must be added
 * to the backtrace).
 *
 * Note: The synthesized variable will assume ownership of any string parameters
 * (type, name).
 *
 * @param 1 the type name of the variable
 * @param 2 the variable name
 * @param 3 the address of the variable
 * @param 4 the offset of the variable, if this is an array element; -1 if not
 * @param 5 the owner of the variable (e.g. the containing struct or array)
 * @return the new variable object if creation succeeded, NULL otherwise
 */
ptrace_variable_t *pmodule_variable_create(const char *, const char *,
    unsigned long long, ssize_t, ptrace_variable_t *);

/**
 * Frees any memory associated with the specified synthesized variable. This
 * should not be called if the variable was added to a frame (and that frame is
 * already in the backtrace (a thread) or will be added to the backtrace).
 */
void pmodule_variable_destroy(ptrace_variable_t *);

/*
 * Variable type and value setters.
 */

/**
 * Sets the variable as a struct type.
 */
void pmodule_variable_set_struct(ptrace_variable_t *);
/**
 * Sets the variable as an array type.
 */
void pmodule_variable_set_array(ptrace_variable_t *);
/**
 * Sets the variable as a function type.
 *
 * @param 2 the symbol of the function; may be NULL
 */
void pmodule_variable_set_function(ptrace_variable_t *, const char *);
/**
 * Sets the variable as a string type.
 *
 * @param 2 the string's symbol; may be NULL
 */
void pmodule_variable_set_string(ptrace_variable_t *, const char *);
/**
 * Sets the variable as an enum type.
 *
 * @param 2 the enum's symbol; may be NULL
 * @param 3 the value of the enum
 */
void pmodule_variable_set_enum(ptrace_variable_t *, const char *, long long);
/**
 * Sets the variable as a reference type.
 *
 * @param 2 the value of the reference
 * @param 3 the reference's symbol; may be NULL
 */
void pmodule_variable_set_ref(ptrace_variable_t *, unsigned long long,
    const char *);
/**
 * Sets the variable as a complex double type.
 */
void pmodule_variable_set_cdbl(ptrace_variable_t *, double complex);
/**
 * Sets the variable as a long complex double type.
 */
void pmodule_variable_set_cldbl(ptrace_variable_t *, long double complex);
/**
 * Sets the variable as a complex float type.
 */
void pmodule_variable_set_cflt(ptrace_variable_t *, float complex);
/**
 * Sets the variable as a double type.
 */
void pmodule_variable_set_dbl(ptrace_variable_t *, double);
/**
 * Sets the variable as a long double type.
 */
void pmodule_variable_set_ldbl(ptrace_variable_t *, long double);
/**
 * Sets the variable as a float type.
 */
void pmodule_variable_set_flt(ptrace_variable_t *, float);
/**
 * Sets the variable as an unsigned long long type.
 */
void pmodule_variable_set_ull(ptrace_variable_t *, unsigned long long);
/**
 * Sets the variable as a long long type.
 */
void pmodule_variable_set_ll(ptrace_variable_t *, long long);

/* Frames */

/**
 * Synthesizes a frame to which a user can add variables and other identifiers.
 * Frame-specific values (e.g. symbol) are not required to be set; they
 * will default to their default/unresolved values.
 *
 * Synthesized frames should be added to a backtrace with
 * pmodule_frame_add_after() in order to have any effect (i.e., to show up
 * in a backtrace view).
 */
ptrace_frame_t *pmodule_frame_create(void);

/**
 * Frees any memory associated with the specified synthesized frame. This
 * should not be called if the frame has already been added to a thread in
 * the backtrace.
 */
void pmodule_frame_destroy(ptrace_frame_t *);

/**
 * Adds the specified variable to the specified frame. Any variables with
 * non-NULL owners will be inserted after their owners, provided the owner
 * is already contained within the frame.
 *
 * @param 1 the frame to which to add the variable
 * @param 2 the variable to add
 * @param 3 true if the variable should be inserted directly after its owner,
 * if it has one (owner should already be in the frame), false otherwise
 * @return true if insertion succeeded, false otherwise
 */
bool pmodule_frame_add_variable(ptrace_frame_t *, ptrace_variable_t *, bool);

/**
 * Inserts a frame into a thread relative to the specified frame.
 *
 * @param 1 the relative frame already in the thread
 * @param 2 the frame to insert
 * @param 3 true if the frame should be inserted after the relative frame; false
 * if it should be inserted before
 * @return true if insertion succeeded, false otherwise
 */
bool pmodule_frame_add_relative(ptrace_frame_t *, ptrace_frame_t *, bool);

/**
 * Inserts a frame at the head or tail of a thread's backtrace.
 *
 * @param 1 the thread to hold the frame
 * @param 2 the frame to insert
 * @param 3 true to insert at the head, false to insert at the tail
 * @return true if insertion succeeded, false otherwise
 */
bool pmodule_thread_add_frame(ptrace_thread_t *, ptrace_frame_t *, bool);

/*
 * Frame value setters.
 *
 * Note: The synthesized frame will assume ownership of any string parameters
 * (type, name).
 */

void pmodule_frame_set_basename(ptrace_frame_t *, const char *);
void pmodule_frame_set_symbol(ptrace_frame_t *, const char *);
void pmodule_frame_set_directory(ptrace_frame_t *, const char *);
void pmodule_frame_set_path(ptrace_frame_t *, const char *);

/**
 * Sets the frame's line number.
 */
void pmodule_frame_set_line(ptrace_frame_t *, unsigned long long);

/**
 * Set's the frame's declaration line number (e.g. the line on which
 * the frame's function was declared).
 */
void pmodule_frame_set_decl_line(ptrace_frame_t *, unsigned long);

/**
 * Extracts the nth C++ template parameter from a type name. Resultant pointer
 * will need to be free()d.
 * 
 * @param 1 the parameter string
 * @param 2 the input type
 * @param 3 the number of argument, 0-indexed (0 for first)
 * @return true for success, false otherwise
 */
bool pmodule_cpp_template_nth_argument(char **, const char *, size_t);

struct bt_type;
struct bt_dwarf;

/**
 * Tries to match a C++ name with a name within the provided DWARF data.
 * 
 * @param 1 DWARF data
 * @param 2 The name
 * @return Pointer to a bt_type object on success, NULL otherwise.
 */
struct bt_type *pmodule_cpp_get_bt_type_from_name(struct bt_dwarf *d,
    const char *name);

/*
 * Internal only function.
 */
void ptrace_module_register(const struct pmodule_internal *);

#ifdef __cplusplus
}
#endif

#endif /* __PTRACE_MODULE_H */
